/*
 * ProtoDemo
 *
 *  Created on: 21.03.2014
 *      Author: vladimir
 */

const int protodemo_major = 1;
const int protodemo_minor = 0;

bool need_trace = false;


#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#else
#include <termios.h>
#include <sys/resource.h>
#include <sys/stat.h>

#endif

#include "box.h"


#include <vector>
#include <stdio.h>
#include <fcntl.h>
#include <string>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>

#ifdef WIN32
typedef HANDLE Port;
#else
typedef int Port;
#endif


enum {
	HPROTO_CMD_PING = 1,

	HPROTO_CMD_BRIDGE = 0xF9,
	HPROTO_CMD_CONFIG = 0xFA,
	HPROTO_CMD_ACCESSCODE = 0xFB,
	HPROTO_CMD_SWITCH_BOOT_APP = 0xFC,
	HPROTO_CMD_DEBUG = 0xFD,
	HPROTO_CMD_UPGRADE = 0xFE,
	HPROTO_CMD_ID = 0xFF,
};

typedef enum {
	HPROTO_OK = 0x00,
	HPROTO_ERROR_INVALID_CMD = 0x01,
	HPROTO_ERROR_INVALID_DATA = 0x02,
	HPROTO_ERROR_INVALID_STATE = 0x03,
	HPROTO_ERROR_NO_ACCESS_CODE = 0x04,
	HPROTO_ERROR_BUSY = 0x05,
} HProtoResult;


const unsigned char hproto_sync = 0x55;
const unsigned char proto_base_size = 7;

typedef struct {
	unsigned char sync;
	unsigned char size;
	unsigned char tmp1;
	unsigned char tmp2;
	unsigned char cmd;
	unsigned char cmd_data[256 - 7];
} Packet;


static const char* GetErrorByResultCode (HProtoResult result) {
	if (result == HPROTO_ERROR_INVALID_CMD)
		return "Invalid command";
	else if (result == HPROTO_ERROR_INVALID_DATA)
		return "Invalid data";
	else if (result == HPROTO_ERROR_INVALID_STATE)
		return "Invalid state";
	else if (result == HPROTO_ERROR_NO_ACCESS_CODE)
		return "No access code (limited)";
	else if (result == HPROTO_ERROR_BUSY)
		return "Busy";

	return "Unknown error";

}

static unsigned int GetCRC16 (const unsigned char* bufData, unsigned int sizeData) {
	const int polynomial = 0x08408;
	unsigned int CRC, i;
	unsigned char j;
	CRC = 0;
	for (i = 0; i < sizeData; i++) {
		CRC ^= bufData[i];
		for (j = 0; j < 8; j++) {
			if (CRC & 0x0001) {
				CRC >>= 1;
				CRC ^= polynomial;
			}
			else CRC >>= 1;
		}
	}
	return CRC;
}

unsigned short CalcCRC (Packet* packet) {
	return GetCRC16 ((const unsigned char*) packet, packet->size - 2);
}

void SetCRC (Packet* packet) {
	*((unsigned short*) ((unsigned char*) packet + packet->size - 2)) =
		CalcCRC (packet);
}

unsigned short GetCRC (Packet* packet) {
	return *((unsigned short*) ((unsigned char*) packet + packet->size - 2));
}


std::string DumpHex (Packet* packet) {
	std::string str;
	unsigned int len = packet->size;
	unsigned char* data = (unsigned char*) packet;
	unsigned int i = 0;
	while (true) {
		if ((i > 0 && !(i % 16)) || i >= len) {
			if (i >= len) {
				for (int j = i; j % 16; j++)
					str += "   ";
			}
			str += "| ";
			static const unsigned char read_chars[] =
				{'+', '-', '=', '_', '/', '\\', '\'', '"', '.', ',', ' '};
			for (unsigned int j = i - ((i - 1) % 16) - 1; j < i; j++) {
				bool is_read_char =
						('a' <= data[j] && data[j] <= 'z') ||
						('A' <= data[j] && data[j] <= 'Z') ||
						('0' <= data[j] && data[j] <= '9');

				for (int k = 0; k < sizeof (read_chars); k++) {
					if (data[j] == read_chars[k]) {
						is_read_char = true;
						break;
					}
				}
				if (is_read_char)
					str += data[j];
				else
					str += ".";
			}

			if (i >= len)
				break;
			str += "\n";
		}
		char buf[8];
		sprintf (buf, "%.2X ", data[i]);
		str += buf;
		i++;
	}
	return str;
}

Packet* ParsePacket (unsigned char** p, unsigned int* size) {
	Packet* packet;
	while (*size > 0) {
		if (**p != hproto_sync) {
			(*p)++; (*size)--;
			continue;
		}
		if (*size < 5)
			return NULL;

		packet = (Packet*) *p;
		if (packet->size > 64) {
			(*p)++; (*size)--;
			continue;
		}
		if (packet->size > *size)
			return NULL;

		if (GetCRC (packet) != CalcCRC (packet)) {
			(*p)++; (*size)--;
			continue;
		}
		(*p) += packet->size;
		(*size) -= packet->size;
		return packet;
	}
	return NULL;
}


bool SendPacket (Port port, Packet* packet) {
	packet->sync = hproto_sync;
	packet->tmp1 = 0;
	packet->tmp2 = 0;

	SetCRC (packet);

	if (need_trace)
		printf ("-> %s\n", DumpHex (packet).c_str ());
#ifdef WIN32
	DWORD bytes = 0;
	if (!WriteFile (port, (char*) packet, packet->size, &bytes, NULL) || bytes != packet->size) {
		printf ("Failed to write to port %d\n", bytes);
		return false;
	}

#else
	if (write (port, (char*) packet, packet->size) != packet->size) {
		printf ("Failed to write to port\n");
		return false;
	}

#endif
	return true;
}

bool RecvData (Port port, char* data, int size) {

	while (size) {

#ifdef WIN32
		DWORD bytes;

		if (!ReadFile (port, (char*) data, size, &bytes, NULL) || bytes == 0) {
			printf ("Failed to read from port\n");
			return false;
		}
#else
		int bytes;
		bytes = read (port, (char*) data, size);
		if (bytes <= 0) {
			printf ("Failed to read from port %d\n", bytes);
			return false;
		}
#endif
		size -= bytes;
		data += bytes;
	}
	return true;
}

static bool CatchSync (Port port, Packet* packet) {
	for (int i = 0; i < 256; i++) {
		if (!RecvData (port, (char*) packet, 1)) {
			printf ("Failed to read sync\n");
			return false;
		}
		if (packet->sync == hproto_sync)
			return true;

	}
	return false;
}

bool RecvPacket (Port port, Packet* packet) {

	if (!CatchSync (port, packet)) {
		printf ("Out of sync\n");
		return false;
	}

	if (!RecvData (port, ((char*) packet) + 1, 3)) {
		printf ("Failed to read header\n");
		return false;
	}

	// Считываем оставшуюся часть
	if (!RecvData (port, ((char*) packet) + 4, packet->size - 4)) {
		printf ("No data in packet\n");
		return false;
	}

	if (need_trace)
		printf ("<- %s\n", DumpHex (packet).c_str ());

	if (GetCRC (packet) != CalcCRC (packet)) {
		printf ("Bad CRC: %X  != %X\n", GetCRC (packet), CalcCRC (packet));
		return false;
	}


	if (packet->cmd == 0) {
		printf ("Error in response: %s\n", GetErrorByResultCode ((HProtoResult) packet->cmd_data[0]));
		return false;
	}

	return true;
}

// Ожидание пакета с CMD, игнорирование других CMD
bool RecvCmdPacket (Port port, Packet* packet, uint8_t cmd) {

	for (int i = 0; i < 10; i++) {
		if (!RecvPacket (port, packet))
			return false;

		if (cmd == packet->cmd)
			return true;

		if (packet->cmd == HPROTO_CMD_BRIDGE) {

			int size = packet->size - 7;
			if (size > 0) {

/*
				char buf[256 - 7 + 1];
				memcpy (buf, packet->cmd_data, packet->size - 7);
				buf[packet->size - 7] = 0;
				printf (buf);
				printf ("\n\n");
*/
			}
		} else
			printf ("Skip unknown pkt %d\n", packet->cmd);
	}
	printf ("CMD timeout\n");
	return false;
}


// Посылка/получение пакета, игнорирование приема "чужих" CMD
bool SendRecvCmdPacket (Port port, Packet* packet, uint8_t cmd) {

	packet->cmd = cmd;

	if (!SendPacket (port, packet))
		return false;


	return RecvPacket (port, packet);
}


void usage () {
	printf (
		"Box proto demo, version %d.%d\n\n"
		"Usage: protodemo <port> [debug]\n"
		"		id\n"
		"		ping\n"
		"		estatus \n"
		"		dispense\n"
		"		recover\n"
		"		upgrade	<firmware path>\n"
		"Automated: http://www.stampsoft.ru\n",
		protodemo_major, protodemo_minor
	);
}

int Upgrade (Port port, std::string filepath);
int Debug (Port port);

int EStatus (Port port, std::vector<std::string> args);
int Dispense (Port port, std::vector<std::string> args);
int Recover (Port port, std::vector<std::string> args);

int main (int argc, const char* argv[]) {
	if (argc < 3) {
		usage ();
		return 1;
	}

	// Открытие порта
	Port port;
#ifdef WIN32
	std::string path = "\\\\.\\";
	path += argv[1];
	port = CreateFile (path.c_str (), GENERIC_READ | GENERIC_WRITE,
		0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (port == INVALID_HANDLE_VALUE) {
		printf ("Failed to open port");
		return 1;
	}


	DCB dcb;
	if (!GetCommState (port, &dcb)) {
		printf ("Failed to get COM state");
		CloseHandle (port);
		return 2;
	}

	dcb.BaudRate = CBR_115200;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;

	dcb.ByteSize = 8;

	if (!SetCommState (port, &dcb)) {
		printf ("Failed to set COM state");
		CloseHandle (port);
		return 4;
	}

	PurgeComm (port, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);
	SetupComm (port,256,256);


#else


	struct rlimit rl;
	if (getrlimit (RLIMIT_CORE, &rl) == 0) {
		rl.rlim_cur = rl.rlim_max;
		if (setrlimit (RLIMIT_CORE, &rl) < 0)
			printf ("Cann't set limit, no core dump\n");
	}


	port = open (argv[1], O_RDWR | O_NOCTTY);
	if (port < 0) {
		perror ("Failed to open port");
		return 1;
	}
	struct termios opt;

	if (tcgetattr (port, &opt) < 0) {
		return 3;
	}

	opt.c_oflag = 0;
	opt.c_lflag = 0;
	opt.c_iflag = IGNPAR;
	opt.c_cflag |= CREAD | CSTOPB | CLOCAL | CS8;
	opt.c_cflag &= ~CBAUD;
	opt.c_cflag |= B115200;


	if (tcsetattr (port, TCSANOW, &opt) < 0) {
		return 2;
	}

	tcflush (port, TCIOFLUSH);

#endif

	// Идентификация устройства и его характеристик
	Packet packet;


	std::string cmd = argv[2];
	int arg = 3;

	bool need_debug = false;
	bool need_bridge = false;

	if (cmd == "debug" && argc >= 3) {
		need_debug = true;
	} else if (cmd == "bridge" && argc >= 3) {
		need_bridge = true;
	} else if (cmd == "trace" && argc >= 3) {
		need_trace = true;
	}

	if (need_bridge || need_debug || need_trace) {
		arg = 4;
		if (argc >= 4)
			cmd = argv[3];
		else
			cmd = "";
	}

	std::vector<std::string> args;
	for (int i = arg; i < argc; i++)
		args.push_back (argv[i]);

	if (cmd == "ping") {
		Packet packet;
		packet.size = proto_base_size;
		if (!SendRecvCmdPacket (port, &packet, HPROTO_CMD_PING)) {
			printf ("Failed to ping device\n");
			return 1;
		}
		printf ("PING %s\n", (packet.cmd == HPROTO_CMD_PING ? "OK" : "FAILED"));
	} else if (cmd == "id") {
		Packet packet;
		packet.size = proto_base_size;
		if (!SendRecvCmdPacket (port, &packet, HPROTO_CMD_ID))
			return 1;


		uint8_t device = packet.cmd_data[0];
		uint8_t major = packet.cmd_data[1];
		uint8_t minor = packet.cmd_data[2];

		printf (
				"ID: device: %X\n"
				"SW version: %d.%d\n", device, major, minor);


	} else if (cmd == "estatus") {
		EStatus (port, args);
	} else if (cmd == "dispense") {
		Dispense (port, args);
	} else if (cmd == "recover") {
		Recover (port, args);
	} else if (cmd == "upgrade" && argc >= 4) {
		Upgrade (port, argv[arg]);
	} else if (cmd != "") {
		printf ("%s\n", cmd.c_str());
		usage ();
		return 1;
	}

	if (need_debug)
		Debug (port);

	return 0;
}

int Upgrade (Port port, std::string filepath) {
	FILE* fp = fopen (filepath.c_str (), "rb");
	if (!fp) {
		printf ("Failed to open file %s: %s\n", filepath.c_str (), strerror (errno));
		return 1;
	}

	fseek (fp, 0, SEEK_END);
	long filesize = ftell (fp);
	if (filesize <= 0 || fseek (fp, 0, SEEK_SET) < 0) {
		printf ("Failed to obtain file size %s: %s\n", filepath.c_str (), strerror (errno));
		return 1;
	}

	printf ("Firmware size: %d\n", filesize);

	if (filesize > 1 * 1024 * 1024) {
		printf ("Invalid firmware file\n");
		fclose (fp);
		return 1;
	}
	uint8_t data[filesize];

	int res = fread (data, 1, filesize, fp);

	if (res != filesize) {
		printf ("Failed to read file: %s, size = %d\n", strerror (errno), res);
		fclose (fp);
		return 1;
	}
	fclose (fp);

	// Переводим в режим BOOT
	Packet packet;
	packet.size = proto_base_size + 4;

	printf ("Switch to boot mode...\n");

	if (!SendRecvCmdPacket (port, &packet, HPROTO_CMD_SWITCH_BOOT_APP)) {
		printf ("[FAILED] Switch to boot mode failed, try as is..\n");
	} else {
		printf ("[OK] Switched to boot mode\n");
	}

	usleep (100000);

	const int bs = 32;

	for (int pos = 0; pos < filesize; pos += bs) {

		Packet packet;
		packet.size = proto_base_size + 4 + bs;


		packet.cmd_data[0] = (pos >> 24) & 0xFF;
		packet.cmd_data[1] = (pos >> 16) & 0xFF;
		packet.cmd_data[2] = (pos >> 8) & 0xFF;
		packet.cmd_data[3] = (pos >> 0) & 0xFF;

		memcpy (&packet.cmd_data[4], &data[pos], bs);

		printf ("Uploading... %02d %%\r", (pos * 100 ) / filesize);

		// Все пустые (0xFF) кроме первого и последнего блока можно пропустить
		bool empty = true;
		for (int i = 0; i < bs; i++) {
			if (packet.cmd_data[4 + i] != 0xFF) {
				empty = false;
				break;
			}
		}

		if (empty && pos != 0 && pos + bs < filesize) {
			//printf ("skip %d\n", pos);
			continue;
		}

		if (!SendRecvCmdPacket (port, &packet, HPROTO_CMD_UPGRADE)) {
			return 1;
		}

	}

	printf ("Upload done 100%     \n");

	return 0;
}

int EStatus (Port port, std::vector<std::string> args) {
	Packet packet;
	packet.size = proto_base_size;
	if (!SendRecvCmdPacket (port, &packet, BOX_HPROTO_CMD_ESTATUS))
		return 1;

	// Получаем отдельные TLV-пакеты статуса
	uint8_t* data = packet.cmd_data;
	size_t rest = packet.size - proto_base_size;

	while (rest >= 3) {
		uint8_t type = data[0];
		uint8_t length = data[1];

		if (type == BOX_HPROTO_ESTATUS_STATE) {
			BoxState state = (BoxState) data[2];
			printf ("Status:[%d] %s\n", state, box_state_hr[state]);
		} else if (type == BOX_HPROTO_ESTATUS_ERROR) {
			uint8_t error_map = data[2];
			if (error_map) {
				printf ("Errors: ");
				for (int e = BOX_ERROR_EMPTY; e < BOX_ERROR_MAX; e++) {
					if (error_map & (1 << e))
						printf ("%s ", box_error_hr[e]);
				}
				printf ("\n");
			}

		} else {
			printf ("Unknown TLV, type = %d, length = %d\n", type, length);
		}

		data += 2 + length;
		rest -= 2 + length;

	}
	return 0;
}

int Dispense (Port port, std::vector<std::string> args) {
	Packet packet;
	packet.size = proto_base_size;

	if (!SendRecvCmdPacket (port, &packet, BOX_HPROTO_CMD_DISPENSE))
		return 1;

	return 0;
}

int Recover (Port port, std::vector<std::string> args) {
	Packet packet;
	packet.size = proto_base_size;

	if (!SendRecvCmdPacket (port, &packet, BOX_HPROTO_CMD_RECOVER))
		return 1;

	return 0;
}

int Debug (Port port) {
	while (true) {
		Packet packet;
		packet.size = proto_base_size;
		if (!SendRecvCmdPacket (port, &packet, HPROTO_CMD_DEBUG))
			return 1;

		int size = packet.size - 7;
		if (size <= 0)
			continue;

//		printf ("-----<%d>-----\n", size);
		char buf[256 - 7 + 1];
		memcpy (buf, packet.cmd_data, packet.size - 7);
		buf[packet.size - 7] = 0;
		printf (buf);
//		printf ("-----<%d>-----\n", size);
		usleep (100);
	}
}
